/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  JacobDatagramSocket (implementation)
    Package: inet
    File:    JacobDatagramSocket.java


*/
package fi.hut.tcm.jacob.tcpip;

import fi.hut.tcm.jacob.tcpip.udp.*;
import fi.hut.tcm.jacob.tcpip.icmp.*;
import fi.hut.tcm.jacob.tcpip.ip.*;
import fi.hut.tcm.jacob.tcpip.arp.*;
import fi.hut.tcm.jacob.tcpip.ethernet.*;

import fi.hut.tcm.jacob.conduits.*;
import java.util.Enumeration;
import java.util.Vector;
import java.net.*;
import java.io.*;

/**
    JacobDatagramSocket
    
    
*/
public final class JacobDatagramSocket
  extends Object {
    
  private final SimpleScheduler scheduler = SimpleScheduler.instance;
  
  public static final int UDP_PROTOCOL_ID = 17; // Protocol field in IP header
  public static final int ICMP_PROTOCOL_ID = 1;
  private long localEthAddr;
  private byte[] localIpAddr = new byte[4];
  private int localIpInt;
  private int localPort;

  private UdpAdaptor master;
  private UdpAdaptor worker;
  private DatagramPacket packet;
  private boolean packetReceived;

  private static final String itblCfgFilename = new String("itbl.cfg");
  private static final String rtblCfgFilename = new String("rtbl.cfg");

  public JacobDatagramSocket(int port, InetAddress addr) 
       throws java.net.SocketException {

    localIpAddr = addr.getAddress();
    localPort = port;

    /*************
     * Instances *
     *************/

    // Init interface table and routing table

    TcpipInterfaceTable ifTable = null;
    TcpipRoutingTable rTable = null;
    try {
      ifTable = new TcpipInterfaceTable(itblCfgFilename);
      rTable = new TcpipRoutingTable(rtblCfgFilename);
    } catch(IOException e) {
      throw new SocketException("I/O Error reading config files");
    } catch(IllegalArgumentException e) {
      throw new SocketException("Syntax error in config files");
    }

    String ifId = null;
    TcpipInterfaceEntry entry = null;
    Enumeration interfaces = ifTable.listEntries();

    localIpInt =
      ((int) localIpAddr[0] & 0xff) << 24 |      
      ((int) localIpAddr[1] & 0xff) << 16 |      
      ((int) localIpAddr[2] & 0xff) <<  8 |      
      ((int) localIpAddr[3] & 0xff) <<  0;

    while (interfaces.hasMoreElements()) {
      entry = (TcpipInterfaceEntry) interfaces.nextElement();
      if (entry.getDefIpAddr() == localIpInt) {
	ifId = entry.getId();
	break;
      }
    }
    if (ifId == null) {
      System.err.println("Unable to get interface");
      System.exit(-1);
    }
    
    localEthAddr = entry.getEthAddr();
    int mtu = entry.getMtu();

    // Ethernet
    
    /*
    // pseudo, machine-internal ethernet:
    PseudoEthernetAdaptor ethAdapt = new PseudoEthernetAdaptor(EthernetAddress.toArray(localEthAddr));
    */

    // Real ethernet:
    Vector filteredIps = new Vector();
    filteredIps.addElement(new IpAddress("130.233.45.26")); // clotam
    filteredIps.addElement(new IpAddress("130.233.45.255")); // broadcast
    filteredIps.addElement(new IpAddress("130.233.195.255")); // broadcast
    EthernetAdaptor ethAdap =
      new EthernetAdaptor( new EthernetAddress(localEthAddr),
			   filteredIps,
			   "/dev/le", 0 );

    // common to real and pseudo-ethernet:

    EthernetHeaderSession ethHdrSes = 
      new EthernetHeaderSession(EthernetHeaderState.instance, localEthAddr);


    // Instances

    try {
      // ARP
      int myIpAddr = localIpInt;   // necessary?
      Arp arpProtocol =
	new Arp(new EthernetAddress(localEthAddr));

      arpProtocol.addIpAddress(new IpAddress(IpAddress.toString(myIpAddr)));
      // END ARP

      // IP
    
      IP ip = new IP(rTable, ifTable);

      // UDP
      UDP udpProtocol = new UDP();

      // ICMP
      Icmp icmpProtocol = new Icmp();
    
      // Master adaptor.
      master = new UdpAdaptor();
    
      /*
	// Stub sessions   - No longer needed in Jacob ver 3.
      TcpipStubSession ipStubSes =
      new TcpipStubSession(null,
			   IpChecksumFromNetMessenger.instance,
			   // IpStubMessenger.instance
			   null
			   );
    TcpipStubSession udpStubSes =
      new TcpipStubSession(IpHeaderToNetMessenger.instance,
			   UdpChecksumFromNetMessenger.instance, 
			   // UdpStubMessenger.instance
			   null
			   );
			   */

    /*********************
     * Conduit structure *
     *********************/

      // ARP
      // No sending yet, here we will connct the IP parts later.
      ethAdap.getSide(Side.A).attach(arpProtocol.getSide(Side.A));
      arpProtocol.getSide(Side.A).attach(ethAdap.getSide(Side.A));
      // END ARP

      // UDP
      udpProtocol.getSide(Side.B).attach(master.getSide(Side.A));

      // Application

      worker = master.connect(localPort);
      if(worker == null) {
	System.out.println("Failed to acquire UDP port");
	System.exit(1);
      }
      // Register with worker that we want the incoming datagrams:
      worker.setMessageAction(this);

      // IP
      ip.addInterface(arpProtocol.getSide(Side.B), ifId);
      ip.addProtocol(udpProtocol.getSide(Side.A), (byte) UDP_PROTOCOL_ID);
      ip.addProtocol(icmpProtocol.getSide(Side.A), (byte) ICMP_PROTOCOL_ID);

    } catch (SideException e) {
      System.err.println("Host " + e + " failed to initialize");
      e.printStackTrace();
      System.exit(2);
    }
  }

  public void send(DatagramPacket p)
       throws java.io.IOException {
	 worker.send(localIpInt,localPort,p);
  }

  public synchronized void receive(DatagramPacket p) 
  throws java.io.IOException {
    while (!packetReceived) {
      try {
	wait();
      } catch (InterruptedException e) {}
    }
    p.setAddress(packet.getAddress());
    p.setPort(packet.getPort());
    p.setData(packet.getData());
    p.setLength(packet.getLength());
    packetReceived = false;
  }

  public synchronized void put(DatagramPacket p) {
    packet = p;
    packetReceived = true;
    notifyAll();
  }

  public byte[] getLocalIpAddr() {
    byte[] newIpAddr = new byte[4];
    System.arraycopy(localIpAddr, 0, newIpAddr, 0, 4);
    return newIpAddr;    
  }

  public void close() {
  }

}; // JacobDatagramSocket



