/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  TcpipRoutingTable (implementation)
    Package: tcpip
    File:    TcpipRoutingTable.java


*/
package fi.hut.tcm.jacob.tcpip;
import java.util.Hashtable;
import java.util.Vector;
import java.util.Enumeration;
import java.io.*;

/**
    TcpipRoutingTable
    
    Returns the next-hop IP based on the destination IP. Contains all the
    routing logic.
    
*/
public class TcpipRoutingTable {
    private TcpipRoutingEntry entries;
    private TcpipRoutingEntry defaultGateway = null;
    protected static final int PARSE_DIGIT = 0;
    protected static final int PARSE_ID = 1;
    protected static final int PARSE_DOT = 2;
    protected static final int PARSE_EOL = 3;
    protected static String GATEWAY_TAG = new String("default");

    public TcpipRoutingTable() {
    //+ TcpipRoutingTable::TcpipRoutingTable( )
    //- TcpipRoutingTable::TcpipRoutingTable( )
    }

    public TcpipRoutingTable(String configFilename)
    throws IOException, IllegalArgumentException {
    //+ TcpipRoutingTable::TcpipRoutingTable(String)
      StreamTokenizer tokenizer;
      FileInputStream cfgFile;
      try {
	cfgFile = new FileInputStream(configFilename);
	tokenizer = new StreamTokenizer(new InputStreamReader(cfgFile));
      } catch (FileNotFoundException e) {
	throw new FileNotFoundException(configFilename);
      }

      // Init tokenizer
      tokenizer.resetSyntax();
      tokenizer.wordChars('0', '9');
      tokenizer.wordChars('a', 'z');
      tokenizer.wordChars('A', 'Z');
      tokenizer.commentChar('#');
      tokenizer.whitespaceChars(0, ' ');
      tokenizer.eolIsSignificant(true);

      int tokmode = PARSE_DIGIT;
      boolean satisfied = false;
      String newId = null;
      int newAddr = 0;
      int newNetworkAddr = 0;
      int newNetmaskAddr = 0;
      int newNexthopAddr = 0;
      int ipCount = 0;
      TcpipRoutingEntry newEntry = null;
      int ipOffs = 0;
      int value;


      // Parse configuration file and create entries
      try {
	while (tokenizer.nextToken() != StreamTokenizer.TT_EOF) {

	  switch (tokmode) {
	    
	  case PARSE_DIGIT:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      try {
		value = new Integer(tokenizer.sval).intValue();
	      } catch (NumberFormatException e) {
		throw new IllegalArgumentException();
	      }

	      if (value > 0xff) {
		throw new IllegalArgumentException();
	      }
	      
	      newAddr += (int) value << (24 - ipOffs);
	      if ((ipOffs += 8) < 32) {
		tokmode = PARSE_DOT;
	      } else {
		switch (ipCount++) {
		case 0: 
		  newNetworkAddr = newAddr;
		  break;
		case 1:
		  newNetmaskAddr = newAddr;
		  break;
		case 2:
		  newNexthopAddr = newAddr;
		  break;
		}
		ipOffs = 0;
		newAddr = 0;
		if (ipCount >= 3) {
		  ipCount = 0;
		  tokmode = PARSE_ID;
		}
	      }
	      satisfied = false;
	    } else if (tokenizer.ttype == StreamTokenizer.TT_EOL) {
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	  case PARSE_DOT:
	    if (tokenizer.ttype == '.') {
	      tokmode = PARSE_DIGIT;
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	    
	  case PARSE_ID:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      newId = new String(tokenizer.sval);
	      
	      // An entry has been parsed
	      newEntry = new TcpipRoutingEntry(newNetworkAddr, 
					       newNetmaskAddr, 
					       newNexthopAddr,
					       newId);
	      addEntry(newEntry);
	      
	      newNetworkAddr = 0;
	      newNetmaskAddr = 0;
	      newNexthopAddr = 0;
	      satisfied = true;
	    tokmode = PARSE_EOL;
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	  
	  case PARSE_EOL:
	    if (tokenizer.ttype == StreamTokenizer.TT_WORD) {
	      if (tokenizer.sval.equals(GATEWAY_TAG)) {
		setDefaultGateway(newEntry);
	      } else {
		throw new IllegalArgumentException();
	      }
	    } else if (tokenizer.ttype == StreamTokenizer.TT_EOL) {
	      newEntry = null;
	      tokmode = PARSE_DIGIT;
	    } else {
	      throw new IllegalArgumentException();
	    }
	    break;
	    
	  }
	} 
      } catch (IOException e) {
	throw new IOException();
      }

      if (!satisfied) {
	throw new IOException();
      }

      cfgFile.close();

      printEntries();

    //- TcpipRoutingTable::TcpipRoutingTable(String)
    }

    public TcpipRoutingEntry get(int destIpAddr) {
    //+ TcpipRoutingTable::get(int)
      TcpipRoutingEntry entry;
      for (entry = entries; entry != null; entry = entry.next) {
	if (entry.network == (destIpAddr & entry.netmask)) {
	  break;
	}
      }
      if (entry == null) {
	entry = defaultGateway;
      }
      return entry;
    //- TcpipRoutingTable::get(int)
    }

    public void addEntry(TcpipRoutingEntry entry) {
    //+ TcpipRoutingTable::addEntry(TcpipRoutingEntry)
      TcpipRoutingEntry cur;
      TcpipRoutingEntry prev = null;
      for (cur = entries; cur != null; cur = cur.next) {
        if (((long) entry.netmask & 0xffffffff) > 
	    ((long) cur.netmask & 0xffffffff)) {
	  break;
	}
	prev = cur;
      }
      if (prev == null) {
	entries = entry;
      } else {
	prev.next = entry;
      }
      entry.next = cur;
    //- TcpipRoutingTable::addEntry(TcpipRoutingEntry)
    }

    public TcpipRoutingEntry setDefaultGateway(TcpipRoutingEntry entry) {
    //+ TcpipRoutingTable::setDefaultGateway(TcpipRoutingEntry)
      TcpipRoutingEntry oldGateway = defaultGateway;
      defaultGateway = entry;
      return oldGateway;
    //- TcpipRoutingTable::setDefaultGateway(TcpipRoutingEntry)
    }

    public Enumeration listEntries() {
    //+ TcpipRoutingTable::listEntries( )
      Vector ents = new Vector();
      TcpipRoutingEntry cur;
      for (cur = entries; cur != null; cur = cur.next) {
	ents.addElement(cur);
      }
      return ents.elements();
    //- TcpipRoutingTable::listEntries( )
    }

    public void printEntries() {
    //+ TcpipRoutingTable::printEntries( )
      TcpipRoutingEntry cur;
      for (cur = entries; cur != null; cur = cur.next) {
	System.err.println();
	cur.print();
	if (cur == defaultGateway) {
	  System.err.println("(default gateway)");
	}
	System.err.println();
      }
    //- TcpipRoutingTable::printEntries( )
    }

}; //TcpipRoutingTable
