/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
  XXX   %Z%%M% version %I% date:%G%
  Component:  Arp
  Package: arp
  File:    Arp.java
*/

package fi.hut.tcm.jacob.tcpip.arp;
import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;
import java.util.Hashtable;

/**
  Arp
  
  ARP (Address Resolution Protocol).<br>
  - Uses Ethernet broadcast to find Ethernet hardware address of any 
  locally connected host with given IP address.<br>
  - Forwards and receives packets to and from the ARP/IP STUB and to the 
  EthernetHeaderSession. <br>
  - Provides an interface for the user to the ARP Table. <br>
  - Basically, just encapsulates the ARP-related Conduits. <br>
  
  See also: ArpTable.
*/
public class Arp
  extends Protocol {

  /**
    Size of ARP-table
  */
  public static final int ArpTableSize = 16;
  
  /**
  */
  public static final int RequestTableSize = 32;
  
  /**
    Size of IP-address
  */
  public static final int SIZEOFIPADDR = 4;
  
  /**
    Size of EtherNet address
  */
  public static final int SIZEOFETHERADDR = 6;
  
  /**
    Size of Ethernet header without preamble.
  */
  public static final int SIZEOFETHERHEADER = 14;
  
  /**
    Arp Table return code: Ethernet address found
  */
  public static final byte ARPOK = 0;
  
  /**
    Arp Table return code: Ethernet address not found
  */
  public static final byte ARPNOTINTABLE = -1;
  
  /**
    Arp Table return code: failure
  */
  public static final byte ARPFAILURE = -2;
  
  /**
    ARP Table return code: Ethernet address found but old
  */
  public static final byte ARPOKBUTOLD = 1;
  
  /**
    Time (milli seconds) how long Arp Table entry 
    (IP address) is concidered unreachable until
    new request can be done.
  */
  public static final long ARPFAILEDTIME = 2*60*1000;
  
  /**
    Time (milli seconds) after which Arp Table entry
    expires.
  */
  public static final long ARPEXPIREDTIME = 30*60*1000;
  
  /**
    Time (milli seconds) after which Arp Table entry
    is still valid but should be checked soon.
  */
  public static final long ARPOKBUTOLDTIME = 20*60*1000;
  
  /**
    Arp Table code: entry is immutable.
  */
  public static final byte ARPIMMUTABLE = 1;
  
  /**
    Arp Table code: entry is mutable.
  */
  public static final byte ARPMUTABLE = 2;
  
  /**
    The different parts of the ARP protocol consult
    this variable for their debug level.The levels
    are as follows:<br>
    <ul>
      <li> 0 - no debug output
      <li> 1 - informational messages
      <li> 2 - detailed debug printout, not for end-users
      <li> 3 - print everything
    </ul>
  */
  public static int DebugLevel = 0;
  
  /**
  */
  private ArpTable arpTable;
  
  /**
  */
  private ArpReplyAdaptor arpReplyAdap;
  
  /**
    ARP-table can't be cloned.
  */
  protected void deepcopy()
    throws CloneNotSupportedException  {
  //+ Arp::deepcopy( )
      throw new CloneNotSupportedException("XXX");
  //- Arp::deepcopy( )
  }
  
  /**
  */
  public Arp(EthernetAddress ethAddr) {
  //+ Arp::Arp(EthernetAddress)
    
    // ARP instances

    arpTable = new ArpTable();

    EthernetHeaderSession ethHdrSes = 
      new EthernetHeaderSession(EthernetHeaderState.instance, ethAddr.toLong());
    EtherAddrAccessor etherAddrAccessor = new EtherAddrAccessor(arpTable);
    Mux etherAddrMux = new Mux(etherAddrAccessor);
    ArpIcmpAdaptor arpIcmpAdapt = new ArpIcmpAdaptor();
    arpReplyAdap = 
      new ArpReplyAdaptor(arpTable, ethAddr);
    ArpRequestSession arpRequestSes = 
      new ArpRequestSession(arpTable, ArpRequestState.instance, ethAddr);
    ArpSelectorAccessor arpSelectorAccessor = new ArpSelectorAccessor();
    Mux arpSelectorMux = new Mux(arpSelectorAccessor);
    Terminator arpSelectorMuxTer = new Terminator(); 

    try {
      
      // ARP Conduit Structure
      
      sides[INSIDE_A].attach(ethHdrSes.getSide(Side.A));
      ethHdrSes.getSide(Side.A).attach(sides[INSIDE_A]);

      ethHdrSes.getSide(Side.B).attach(arpSelectorMux.getSide(Side.A));
      arpSelectorMux.getSide(Side.A).attach(ethHdrSes.getSide(Side.B));

      Side arpSeleB1 = arpSelectorMux.createSide("1");
      Side arpSeleB2 = arpSelectorMux.createSide("2");
      Side arpSeleB3 = arpSelectorMux.createSide("3");

      Side etherAddrB1 = etherAddrMux.createSide("1");
      Side etherAddrB2 = etherAddrMux.createSide("2");

      etherAddrB1.attach( arpSeleB1 );
      arpSeleB1.attach( etherAddrB1 );

      arpReplyAdap.getSide(Side.A).attach( arpSeleB2 );
      arpSeleB2.attach(arpReplyAdap.getSide(Side.A));

      arpRequestSes.getSide(Side.A).attach( arpSeleB3 );
      arpSeleB3.attach( arpRequestSes.getSide(Side.A));

      arpSelectorMux.getSide(Side.B).attach( arpSelectorMuxTer.getSide(Side.A));
      arpSelectorMuxTer.getSide(Side.A).attach(arpSelectorMux.getSide(Side.B));

      etherAddrB2.attach(arpRequestSes.getSide(Side.B));
      arpRequestSes.getSide(Side.B).attach(etherAddrB2);

      etherAddrMux.getSide(Side.B).attach(arpIcmpAdapt.getSide(Side.A));
      arpIcmpAdapt.getSide(Side.A).attach(etherAddrMux.getSide(Side.B));

      sides[INSIDE_B].attach(etherAddrMux.getSide(Side.A));
      etherAddrMux.getSide(Side.A).attach(sides[INSIDE_B]);
    } catch (SideException e) {
      System.err.println("Host " + e + " failed to initialize");
      System.exit(2);
    }


  //- Arp::Arp(EthernetAddress)
  }
  
  /**
  */
  public void addIpAddress(IpAddress ipAddr) {
  //+ Arp::addIpAddress(IpAddress)
    arpReplyAdap.addIpAddress(ipAddr);
  //- Arp::addIpAddress(IpAddress)
  }
  
  /**
  */
  public void removeIpAddress(IpAddress ipAddr) {
  //+ Arp::removeIpAddress(IpAddress)
    arpReplyAdap.removeIpAddress(ipAddr);
  //- Arp::removeIpAddress(IpAddress)
  }
  
} // Arp

// +*************** UNUSED CODE ***************

// -*************** UNUSED CODE ***************
