/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
  XXX   %Z%%M% version %I% date:%G%
  Component:  ArpSelectorAccessor
  Package: arp
  File:    ArpSelectorAccessor.java
*/

package fi.hut.tcm.jacob.tcpip.arp;
import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;

/**
  ArpSelectorAccessor
  
  Used by ARPSelectorMUX to multiplex packets coming from Ethernet.  <br>
  - Routes ARP requests, ARP replies and IP packets from the network to the
   appropriate conduits according to
       the frame type field in the Ethernet header and IP header type in 
  the case of Arp request or reply. 
       Unrecognized packets are sent to the ARP selector terminator. <br>
  - Marks packets going to the network in the frame type field and forwards
   them to the Ethernet Adaptor. 
   
*/
public class ArpSelectorAccessor
  extends Object
  implements Accessor {

  /**
    Returns key for multiplexing packets from Ethernet: <br>
    "1" IP packet multiplexed to EtherAddMUX <br>
    "2"  ARP REQUEST multiplexed to ARPReplyAdaptor <br>
    "3" ARP REPLY multiplexed to ArpRequestSession <br>
    
    @param    message
    @return    Object    returns the corresponding key
  */
  public Object getKey(Message message) {
  //+ ArpSelectorAccessor::getKey(Message)

      Object buf = message.getContents();
      TcpipOutOfBand oob = (TcpipOutOfBand) message.getOob();

      int arpType;

      if (Arp.DebugLevel > 2) {  // #Debug#
	System.out.println("ArpSelectorAccessor:EthFrameType seems to be " +
			   oob.getEthFrameType() + " decimal, " +
			   Integer.toHexString( oob.getEthFrameType() ) +
			   " hexadecimal.");
      }
      
      if(oob.getEthFrameType() ==  0x0800) {
	if (Arp.DebugLevel > 2) {  // #Debug#
	  System.out.println("ArpSelectorAccessor detected IP package");
	}
	message.setMessenger(null);
	return "1";
      }
      
      if ((oob.getEthFrameType() ==  0x0806)
	  || (oob.getEthFrameType() == 0xef2a)) {
	
	arpType = UtilMsgr.i.getShortUnaligned(buf, 6);
	if (arpType == 0x0001) { // ARP REQ = 0x0001
	  if (Arp.DebugLevel > 2) {  // #Debug#
	    System.out.println("ArpSelectorAccessor detected ARP Request");
	  }
	  return "2"; //ARP REQUEST
	}

	if (arpType == 0x0002) { // ARP ANS = 0x0002
	  // Remember to set the Messenger ...
	  message.setMessenger(ArpRequestResponseMessenger.instance);
	  if (Arp.DebugLevel > 2) {  // #Debug#
	    System.out.println("ArpSelectorAccessor detected ARP Response");
	  }
	  return "3"; //ARP RESPONSE
	}
	System.err.println("ArpSelectorAccessor.getKey: Unknown ARP type");
      } else {
	System.err.print("ArpSelectorAccessor.getKey: Unknown EtherType:");
	System.err.println(oob.getEthFrameType());
      }
      return null; //default

  //- ArpSelectorAccessor::getKey(Message)
  }
  
  /**
    Sets the key for a message.
    
    @param    message
    @param    key
    
  */
  public void setKey(Message message, Object key) {
  //+ ArpSelectorAccessor::setKey(Message, Object)

      if (Arp.DebugLevel > 2) {  // #Debug#
	TcpipOutOfBand oob = (TcpipOutOfBand) message.getOob();
	System.out.println("ArpSelectorAccessor.setKey: Doing nothing, " +
			   "but noticed that EthFrameType = " +
			   oob.getEthFrameType() + " decimal, " +
			   Integer.toHexString(oob.getEthFrameType() ) +
			   " hexadecimal.");
      }

  //- ArpSelectorAccessor::setKey(Message, Object)
  }
  
  /**
  */
  public void unsetKey(Message message) {
  //+ ArpSelectorAccessor::unsetKey(Message)
    // nothing to do here...
  //- ArpSelectorAccessor::unsetKey(Message)
  }
  
} // ArpSelectorAccessor

// +*************** UNUSED CODE ***************
// -*************** UNUSED CODE ***************
