/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  Icmp (implementation)
    Package: icmp
    File:    Icmp.java


*/
package fi.hut.tcm.jacob.tcpip.icmp;
import fi.hut.tcm.jacob.conduits.*;

/**
    Icmp
    
    <p>The Internet Control Message Protocol. This protocol is a required
    part of any IP implementation, even though it logically is a higher
    level protocol, using IP as its transport.<br>
    This conduit encapsulates the other ICMP modules and effectively
    acts as an interface for other protocols wanting to use the ICMP
    protocol.
*/
public class Icmp
  extends Protocol
  implements Debuggable
{
    /**
        Specifies the byte offset of the <code>ICMP</code> type in the
        <code>ICMP</code> message.
    */
    public static int TYPE_OFFSET = 0;
    /**
        Specifies the byte offset of the <code>ICMP</code> code in the
        <code>ICMP</code> message.
    */
    public static int CODE_OFFSET = 1;
    /**
        Specifies the byte offset of the <code>ICMP</code> checksum in the
        <code>ICMP</code> message.
    */
    public static int CHECKSUM_OFFSET = 2;
    /**
        The debug output level (0, 1, 2 or 3). Level 0 is no debug output,
        whereas level 3 prints the most debug output. Level 1 is intended
        to be informational,
        levels 2 and 3 give detailed debug information.
    */
    public static int DebugLevel = 0;

    /**

    */
    public Icmp() {

      // Instances
      IcmpEchoAnswerAdaptor answAdap =
	new IcmpEchoAnswerAdaptor();
      IcmpApplicationAccessor applAcce =
	new IcmpApplicationAccessor();
      Mux applMux =
	new Mux(applAcce);
      IcmpHeaderSession headerSessProto =
	new IcmpHeaderSession(IcmpHeaderState.instance);
      Side[] factProtoSides = new Side[2];
      factProtoSides[Side.A] = applMux.getSide(Side.B);
      factProtoSides[Side.B] = null;  // The installer will know this side.
      ConduitFactory applFact =
	new ConduitFactory(headerSessProto, factProtoSides);
      IcmpTypeAccessor typeAcce =
	new IcmpTypeAccessor();
      Mux typeMux =
	new Mux(typeAcce);
      Terminator typeMuxTerm =
	new Terminator(); 
      Terminator applMuxTerm =
	new Terminator();
      Terminator factTerm =
	new Terminator();
      IcmpChecksumSession sumSess = 
	new IcmpChecksumSession(IcmpChecksumState.instance);

      // Define the conduit structure of the system

      try {
	sides[INSIDE_B].attach(applFact.getSide(Side.B));
	applFact.getSide(Side.B).attach(sides[INSIDE_B]);

	applFact.getSide(Side.A).attach(factTerm.getSide(Side.A));
	applMux.getSide(Side.B).attach(applMuxTerm.getSide(Side.A));

	applMux.getSide(Side.A).attach(typeMux.createSide(new Integer(0)));
	answAdap.getSide(Side.A).attach(typeMux.createSide(new Integer(8)));
	
	typeMux.getSide(Side.A).attach(sumSess.getSide(Side.B));
	typeMux.getSide(Side.B).attach(typeMuxTerm.getSide(Side.A));
	
	sumSess.getSide(Side.A).attach(sides[INSIDE_A]);
	sumSess.getSide(Side.B).attach(typeMux.getSide(Side.A));
	
	sides[INSIDE_A].attach(sumSess.getSide(Side.A));
	
      } catch (SideException e) {
	System.out.println("Icmp failed to initialize: " + e);
	System.exit(2);
      }

      // Set a sensible default messenger, as it doesn't make sense
      // that objects external to the protocol would have to know
      // about the internals of the protocol, i.e. which
      // conduit will be encountered first and what kind of messenger
      // it needs.
      
      (sides[Side.A]).setMessenger(IcmpChecksumFromNetMessenger.instance);

    }

  public Object clone()
       throws CloneNotSupportedException
  {
    Conduit c;
    Protocol clone = (Protocol)super.clone();
    // Get a reference to the ConduitFactory.
    c = getConduitCheckConnection(clone.getSide(INSIDE_B));
    if(!(c instanceof ConduitFactory))
      System.out.println("Assertion failed in Icmp.clone()");
    ConduitFactory applFact = (ConduitFactory)c;
    // Get a reference to the application mux.
    c = getConduitCheckConnection(clone.getSide(INSIDE_A));
    c = getConduitCheckConnection(c.getSide(Side.B));
    c = getConduitCheckConnection(c.getSide(2));
    if(!(c instanceof Mux))
      System.out.println("Assertion failed (3)");
    Mux applMux = (Mux)c;
    applFact.setEndpoint(Side.A,applMux.getSide(Side.B));
    applFact.setEndpoint(Side.B,null); // The installer will know this side.
    return (Object)clone;
  }
    
  private Conduit getConduitCheckConnection(Side side)
  {
    if(side == null)
      System.out.println("check connection failed (" + 1 + ")");
    Side side2 = side.getAttached();
    if(side2 == null)
      System.out.println("check connection failed (" + 2 + ")");
    if(side2.getAttached() != side)
      System.out.println("check connection failed (" + 3 + ")");
    Conduit conduit = side2.getConduit();
    if(conduit == null)
      System.out.println("check connection failed (" + 4 + ")");
    return conduit;
  }

  public void debug()
  {
    System.out.println("DUMP: Checking protocol structure...");

    if(!(this instanceof Protocol))
      System.out.println("Assertion failed (1)");

    if(sides[Side.A].getAttached() != null)
      System.out.println("DUMP: Side A attached to something");
    if(sides[Side.B].getAttached() != null)
      System.out.println("DUMP: Side B attached to something");

    dump();

    Conduit c;
    int numSides;

    // Factory from protocol.
    c = getConduitCheckConnection(sides[INSIDE_B]);
    c.dump();
    // Terminator from factory.
    getConduitCheckConnection(c.getSide(Side.A)).dump();

    // ChecksumSession from protocol.
    c = getConduitCheckConnection(sides[INSIDE_A]);
    c.dump();
    // Type mux from ChecksumSession.
    c = getConduitCheckConnection(c.getSide(Side.B));
    c.dump();
    if(!(c instanceof Mux))
      System.out.println("Assertion failed (2)");
    numSides = c.getSideCount();
    System.out.println("DUMP: There are currently " + (int)(numSides-2)
		       + " types");
    // Terminator from type mux.
    getConduitCheckConnection(c.getSide(Side.B)).dump();
    // Answer adaptor from type mux.
    getConduitCheckConnection(c.getSide(3)).dump();
    // Application mux from type mux.
    c = getConduitCheckConnection(c.getSide(2));
    c.dump();
    if(!(c instanceof Mux))
      System.out.println("Assertion failed (3)");
    // Terminator from application mux.
    getConduitCheckConnection(c.getSide(Side.B)).dump();
    // All header sessions from application mux.
    numSides = c.getSideCount();
    System.out.println("DUMP: There are currently " +
		       (int)(numSides-2) + " applications");
    for(int i=2; i<numSides; i++)
      getConduitCheckConnection(c.getSide(i)).dump();
  }


  /**
    Creates a deep copy of this protocol by replacing all instance
    variables with references to <bold>copies</bold> of their contents.
    */
  public void deepcopy() {
    //+ Icmp::deepcopy( )
    //- Icmp::deepcopy( )
  }

}; //Icmp



