/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  IcmpApplicationAdaptor (implementation)
    Package: icmp
    File:    IcmpApplicationAdaptor.java


*/
package fi.hut.tcm.jacob.tcpip.icmp;
import fi.hut.tcm.jacob.conduits.SimpleScheduler;
import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;

/**
    IcmpApplicationAdaptor
    
    <p>Creates ICMP ECHO REQUEST messages and receives echo replies.
*/
public class IcmpApplicationAdaptor
  extends Adaptor {

    private ReplyListener replyListener = null;

    public IcmpApplicationAdaptor() {
      // Note that we cannot do the connect() stuff here:
      // This adaptor is not yet attached to the Icmp protocol!
    }

    public void setReplyListener(ReplyListener listener)
    {
      replyListener = listener;
    }
    
    public /*synchronized*/ IcmpApplicationAdaptor connect() {
      // We'll have to send an Installer to the ICMP protocol,
      // requesting service.

      IcmpApplicationAdaptor workerAdaptor = new IcmpApplicationAdaptor();

      Side[] oobsides = new Side[2];
      oobsides[Side.A] = null;   // Factory/mux fills in this side
      oobsides[Side.B] = workerAdaptor.getSide(Side.A);
      InstallOob ioob = new InstallOob( oobsides, null);  // no prototype
      // The message content is ours to choose, so we include
      // a reference to the workerAdaptor so that we know for
      // which adaptor the session was created when we receive
      // installation complete notification.
      Message msg =
      	Message.newMessage((Object)workerAdaptor, Installer.instance, ioob);
      if (sides[Side.A] != null) {
	MessageTransporter mt =
	  MessageTransporter.newMessageTransporter(this.sides[Side.A].
						   getAttached(), 
						   msg);
	SimpleScheduler.instance.makeRunnable(mt);
	Thread.currentThread().yield();
      }
      // Wait for a reply before
      // returning "OK" to the caller.
      workerAdaptor.waitConnection(true);
      return workerAdaptor;
    }

  public /*synchronized*/ void disconnect(IcmpApplicationAdaptor worker)
  {
    // We'll have to send an Uninstaller to the ICMP protocol,
    // requesting service.
    //
    UninstallOob uioob = new UninstallOob((Conduit)worker);
    Message msg = Message.newMessage(worker, Uninstaller.instance, uioob);
    if(sides[Side.A] != null) {
      MessageTransporter mt =
	MessageTransporter.newMessageTransporter(this.sides[Side.A].
						 getAttached(),msg);
      SimpleScheduler.instance.makeRunnable(mt);
      Thread.currentThread().yield();
    }
    // Wait for a reply before returning.
    worker.waitConnection(false);
  }

  private boolean isConnected = false;
  public synchronized void setConnected(boolean is) {
    isConnected = is;
    notifyAll();
  }
  public synchronized void waitConnection(boolean is) {
    while(is != isConnected) {
      try {
	wait();
      } catch(InterruptedException e) {
	System.out.println("wait() interrupted");
      }
    }
  }

    /** Prints out the status if the message passed in is
      a notification message.
      */
  protected boolean filter(Message message)
  {
    //+ IcmpApplicationAdaptor::filter(Message)
    Messenger messenger = message.getMessenger();
    if(messenger instanceof NotifyMessenger) {
      Message request = ((NotifyMessenger)messenger).getRequest(message);
      Messenger reqMsgr = request.getMessenger();
      if(reqMsgr instanceof InstallMessenger) {
	IcmpApplicationAdaptor worker =
	  (IcmpApplicationAdaptor)request.getContents();
	if(worker == null)
	  System.out.println("Assertion failed in IcmpApplicationAdaptor." +
			     "filter()");
	worker.setConnected(true);
	if(Icmp.DebugLevel > 1)
	  System.out.println("Got installation notification, status " +
			     ((NotifyOob)message.getOob()).getStatusCode());
      } else if(reqMsgr instanceof UninstallMessenger) {
	IcmpApplicationAdaptor worker =
	  (IcmpApplicationAdaptor)request.getContents();
	if(worker == null)
	  System.out.println("Assertion failed in IcmpApplicationAdaptor." +
			     "filter()");
	worker.setConnected(false);
	if(Icmp.DebugLevel > 1)
	  System.out.println("Got uninstallation notification, status " +
			     ((NotifyOob)message.getOob()).getStatusCode());
      }
      return true;
    }
    return false;
    //- IcmpApplicationAdaptor::filter(Message)
  }

    /**
        Receives replies to echo requests.
        
        @param message The <code>Message</code> carrying the echo reply
    */
    public /*synchronized*/ void adapt(Message message) {
    //+ IcmpApplicationAdaptor::adapt(Message)

      if(Icmp.DebugLevel > 3)
	System.out.println("Now in IcmpApplicationAdaptor.adapt()");

      // Incoming replies to packets we have sent should arrive here.

      if(filter(message)) return;

      Object buf = message.getContents();
      TcpipOutOfBand oob = (TcpipOutOfBand) message.getOob();
      int len = IcmpUtilityMessenger.instance.getLength(buf);
      
      int src = oob.getSrcIpAddr();
      int dst = oob.getDestIpAddr();
      int ident = IcmpUtilityMessenger.instance.getShort(buf, 2);  // cms: updated
      int seq = IcmpUtilityMessenger.instance.getShort(buf, 3);    // cms: updated
      
   
      String txt = "<-- Echo Reply: src = " + IpAddress.toString(src) +
			 ", dst = " + IpAddress.toString(dst);
    
      if(replyListener != null)
        replyListener.pong(txt);
      else {
        System.out.println(txt);
        System.out.println("    ident = " + ident + ", seq = " + seq);
      }

      if (Icmp.DebugLevel > 1) {  // #Debug#
        System.out.println("    Packet dump:");
        // hexDump(buf);
        System.out.println("hexDump not supported yet");
      }
      
    //- IcmpApplicationAdaptor::adapt(Message)
    }

    /**
        Send an ICMP ECHO REQUEST message.
        
        @param    dstIP  destination IP
        @param    ident  packet identfier
        @param    seq  packet sequence number
    */
    public void send(int dstIp, int seq) {
    //+ IcmpApplicationAdaptor::send(int, int)

    // Maximum size for IP header is 60 octets. (RFC 791).
    // Allow for that, plus some extra header room for ethernet
    // and unforseen requirements. = total of 96 octets.
    // ICMP header for Echo messages is 8 octets + optional data (we use 24)
      // Variables: newBinaryMessage(MAXSIZE, INITIALOFFSET, INITIALLENGTH);
      Object buf =
	BufferMessenger.newBuffer(256, 96, 32);

      IcmpUtilityMessenger.instance.setByte(buf, Icmp.TYPE_OFFSET, 8);
      IcmpUtilityMessenger.instance.setByte(buf, Icmp.CODE_OFFSET, 0);

      // data[2] and data[3] are the checksum fields

      //     IcmpUtilityMessenger.instance.setShort(buf, 2, ident);  // cms: updated
      IcmpUtilityMessenger.instance.setShort(buf, 3, seq);  // cms: updated
      IcmpUtilityMessenger.instance.setInt(buf, 2, 0x01020304); // cms: updated these also
      IcmpUtilityMessenger.instance.setInt(buf, 3, 0x05060708);
      IcmpUtilityMessenger.instance.setInt(buf, 4, 0x090a0b0c);
      IcmpUtilityMessenger.instance.setInt(buf, 5, 0x0d0e0f10);
      IcmpUtilityMessenger.instance.setInt(buf, 6, 0x11121314);
      IcmpUtilityMessenger.instance.setInt(buf, 7, 0x15161718);
      // buf.setByte(32, 0x19);
      
      TcpipOutOfBand oob = new TcpipOutOfBand();
      oob.setDestIpAddr(dstIp);

      oob.setSrcIpAddr(0);  // Let the IP stack figure out our address.

      Message msg =
	Message.newMessage(buf, IcmpHeaderToNetMessenger.instance, oob);

      if (sides[Side.A] != null) {
	MessageTransporter mt =
	  MessageTransporter.newMessageTransporter(sides[Side.A].getAttached(),
						   msg);
	SimpleScheduler.instance.makeRunnable(mt);
      }
    //- IcmpApplicationAdaptor::send(int, int)
    }

}; //IcmpApplicationAdaptor

