/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */

/**
    Copyright (C) 1997 Pekka Nikander & Juha Prssinen.
    All rights Reserved.

    XXX   %Z%%M% version %I% date:%G%

    Module:  IcmpChecksumMessenger (implementation)
    Package: icmp
    File:    IcmpChecksumMessenger.java


*/
package fi.hut.tcm.jacob.tcpip.icmp;
import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;

/**
    IcmpChecksumMessenger
    
    <p>An abstract class that provides the checksum algorithm for
    IcmpChecksumFromNetMessenger and IcmpChecksumToNetMessenger.
    The checksum algorithm calculates the ICMP checksum based on the packet
    header.
*/
public abstract class IcmpChecksumMessenger
  extends BufferMessenger
  implements Messenger {

    /**
        Calculates the checksum.
                
        @param message    the <code>Message</code> for which to calculate
        the checksum
        @return    returns the checksum
    */
    protected int calculateChecksum(Message message) {
    //+ IcmpChecksumMessenger::calculateChecksum(Message)

      Object buf =  message.getContents();

      int len = getLength(buf); 

      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.err.println("Calculating ICMP checksum, packet dump:");
	//hexDump(buf);
	System.err.println("hexDump not yet implemented");
      }

      int elem;
      // optimization: one's complement of the checksum field,
      int csum = 0;
      // calculate the first one by hand:
      //elem = (((int) data[0] & 0xff) << 8) + ((int) data[1] & 0xff);
      elem = getShort(buf, 0);
      csum += elem;
      // loop over the rest of the message:
      for (int i = 4; i < (len / 2) * 2; i += 2) {
	//elem = (((int) data[i] & 0xff) << 8) + (((int) data[i+1] & 0xff));
	elem = getShort(buf, i/2);
	csum += elem;
      }
      if ((len % 2) == 1) { // odd size
	//elem = (((int) data[len-1] & 0xff) << 8);
	elem = getByte(buf, len-1);
	csum += elem;
      }
      while ((csum >> 16) > 0) {
	csum = ((csum >> 16) & 0xffff) + (csum & 0xffff);
      }

      // final one's complement:
      csum = csum ^ 0xffff;

      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("Calculated ICMP checksum = " + csum); // XXX Debug
      }

      return csum;
    //- IcmpChecksumMessenger::calculateChecksum(Message)
    }

    /**
        Gets the checksum for a <code>Message</code> containing one.
                
        @param message which <code>Message</code>
        
        @return    the checksum
    */
    protected int getChecksum(Message message) {
    //+ IcmpChecksumMessenger::getChecksum(Message)
      int csum;
      Object buf = message.getContents();
      csum = getShort(buf, 1);  // remember index 1 = byte position 2
      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("getChecksum: " + csum);
      }
      return csum;
    //- IcmpChecksumMessenger::getChecksum(Message)
    }

    /**
        Adds the ICMP checksum to a message.
                
        @param message  the <code>Message</code>
        @param csum    the checksum to add
    */
    protected void setChecksum(Message message, int csum) {
    //+ IcmpChecksumMessenger::setChecksum(Message, int)
      Object buf = message.getContents();

      setShort(buf, 1, csum);  // remember: index 1 => byte position 2

      if (Icmp.DebugLevel > 1) {  // #Debug#
	System.out.println("setChecksum: " + csum);
      }
    //- IcmpChecksumMessenger::setChecksum(Message, int)
    }

}; //IcmpChecksumMessenger

