/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


import java.util.Random;

import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;
import fi.hut.tcm.jacob.tcpip.icmp.*;


/** Applications usually run in separate processes.
    Well, this one runs in a separate thread.
*/
public class Application
  implements Runnable
{

  private static int globalAppNum = 0;
  private int appNum;

  private IcmpApplicationAdaptor master = null;
  private IcmpApplicationAdaptor worker = null;

  private Thread runner = null;

  private int localIpAddress;
  private int remoteIpAddress;

  private Random rng;

  public Application(IcmpApplicationAdaptor master,
		     int localIpAddress)
    {
      this.master = master;
      this.localIpAddress = localIpAddress;
      rng = new Random();
      appNum = ++globalAppNum;
    }

  public void setRemoteIpAddress(int remoteIpAddress)
  {
    this.remoteIpAddress = remoteIpAddress;
  }

  /** Requests the master adaptor 'master' to create a worker adaptor,
      and to connect the worker to the protocol.
      If 'master' is passed in as null, uses property.
      Saves a reference to the worker in 'worker'.
  */
  public void connect(IcmpApplicationAdaptor master)
    {
      if(master == null)
        master = this.master;

      worker = master.connect();
      if(worker == null) {
        System.out.println("Unable to connect to Icmp Protocol");
        System.exit(1);
      }
      
      worker.setReplyListener(new ReplyListener() {
        public void pong(String txt) { System.out.println("App "+appNum+
							  ": Pong: "+txt);
	gotAck(); }
      });

    }

  /** Takes down connections to the worker.
   */
  public void disconnect()
  {
    master.disconnect(worker);
  }
  
  /** Pings the local machine.
      Sees to it that when a reply is received,
      o.pong() is called.
  */
  public void ping(int ipAddress)
    {
      worker.send(ipAddress, 0);
      Thread.currentThread().yield();
    }

  /** Creates a new thread, and has it call the run() method
      of this class. */
  public void start()
    {
      if(runner != null) {
	System.out.println("start() should only be called once per instance");
	return;
      }
      runner = new Thread(this);
      runner.start();
    }
  
  /** Causes the thread to sleep for 'amount' milliseconds.
   */
  public void sleep(int amount)
    {
      try {
	Thread.sleep(amount);
      } catch (InterruptedException e){ 
	// do nothing.
      }
    }

  /** This method first calls connect(), then
      does some calls to ping(), and finally calls
      disconnect().
  */
  public void run()
    {
      connect(null);
      //sleep(1000);
      for(int i=0; i<10; i++)
	if(rng.nextBoolean()) {
	  System.out.println("App " + appNum + ": Ping " + i);
	  ping(remoteIpAddress);
	  needAck();
	}
      // Wait here until all replies have been received,
      // then disconnect from the network and return.
      if(Icmp.DebugLevel>0)
	System.out.println("App " + appNum + " waiting...");
      waitAcks();
      if(Icmp.DebugLevel>0)
	System.out.println("App " + appNum + " has all acks.");
      disconnect();
      System.out.println("App " + appNum + " done.");
      amFinished();
    }

  private boolean finished = false;
  public synchronized void waitTermination()
  {
    while(!finished) {
      try {
	wait();
      } catch(InterruptedException e) {
	System.out.println("wait() interrupted");
      }
    }
  }
  private synchronized void amFinished()
  {
    finished = true;
    notifyAll();
  }

  private int acksNeeded = 0;
  private synchronized void needAck()
  {
    acksNeeded++;
  }
  private synchronized void gotAck()
  {
    acksNeeded--;
    notifyAll();
  }
  private synchronized void waitAcks()
  {
    while(acksNeeded > 0) {
      try {
	wait();
      } catch(InterruptedException e) {
	System.out.println("wait() interrupted");
      }
    }
  }

  protected void finalize()
  {
    System.out.println("Thread " + appNum + " garbage collected");
  }

}
