/*
  Copyright (c) 1998 Helsinki University of Technology.
  All rights reserved.

  Redistribution and use in source and binary forms are permitted
  provided that the above copyright notice and this paragraph are
  duplicated in all such forms and that any documentation,
  advertising materials, and other materials related to such
  distribution and use acknowledge that the software was developed
  by Helsinki University of Technology, Espoo.  The name of the
  University may not be used to endorse or promote products derived
  from this software without specific prior written permission.
  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
  IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
  WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */


import java.util.Random;

import fi.hut.tcm.jacob.conduits.*;
import fi.hut.tcm.jacob.tcpip.*;
import fi.hut.tcm.jacob.tcpip.udp.*;


/** Applications usually run in separate processes.
    Well, this one runs in a separate thread.
*/
public class Application
  implements Runnable
{

  private static int globalAppNum = 0;
  private int appNum;

  private UdpApplicationAdaptor master = null;
  private UdpApplicationAdaptor worker = null;

  private Thread runner = null;

  // These two are needed by the checksum session.
  private int localIpAddress;
  private int remoteIpAddress;

  private int localPort;
  private int remotePort;

  private Random rng;

  public Application(UdpApplicationAdaptor master,
		     int localIpAddress,
		     int localPort)
  {
    this.master = master;
    this.localIpAddress = localIpAddress;
    this.localPort = localPort;
    rng = new Random();
    appNum = ++globalAppNum;
  }

  public void setDestination(int remoteIpAddress, int remotePort)
  {
    this.remoteIpAddress = remoteIpAddress;
    this.remotePort = remotePort;
  }
  
  /**
    Requests the master adaptor 'master' to create a worker adaptor,
    and to connect the worker to the protocol.
    If 'master' is passed in as null, uses property.
    Saves a reference to the worker in 'worker'.
    */
  public void connect(UdpApplicationAdaptor master)
  {
    if(master == null)
      master = this.master;
    
    worker = master.connect(localPort);
    if(worker == null) {
      System.out.println("App "+appNum+": Connect failed");
    } else {
      worker.setMessageAction(new MessageHandler());
    }
  }

  /** Takes down the connection to the worker.
   */
  public void disconnect()
  {
    if(!master.disconnect(worker))
      System.out.println("App "+appNum+": Disconnect failed");
  }
  



  /**
     Sends a message to the given port of the given remote machine.
     A sequence number can also be given if needed.
  */
  public void send(int remoteIpAddress, int remotePort, int seq)
    {
      worker.send(localIpAddress,localPort,remoteIpAddress,remotePort,
		  appNum,seq);
      Thread.currentThread().yield();
    }





  /** This method first calls connect(), then sends some
      messages to the destination IP address and port, and finally calls
      disconnect().
      */
  public void run()
  {
    connect(null);
    System.out.println("App "+appNum+" connected to the net");
    
    for(int i=0, j=0; i<10; i++)
      if(rng.nextBoolean()) {
	System.out.println("App " + appNum + ": Send " + j);
	send(remoteIpAddress,remotePort,j);
	needAck();
	j++;
      }
    
    // Wait here until all replies have been received,
    // then disconnect from the network and return.
    if(Debugger.level>0)
      System.out.println("App " + appNum + " waiting...");
    waitAcks();
    if(Debugger.level>0)
      System.out.println("App " + appNum + " has all acks.");
    
    disconnect();
    System.out.println("App " + appNum + " done.");
    amFinished();
  }



  /** Creates a new thread, and has it call the run() method
      of this class. */
  public void start()
    {
      if(runner != null) {
	System.out.println("start() should only be called once per instance");
	return;
      }
      runner = new Thread(this);
      runner.start();
    }
  


  private boolean finished = false;
  public synchronized void waitTermination()
  {
    while(!finished) {
      try {
	wait();
      } catch(InterruptedException e) {
	System.out.println("wait() interrupted");
      }
    }
  }
  private synchronized void amFinished()
  {
    finished = true;
    notifyAll();
  }

  private int acksNeeded = 0;
  private synchronized void needAck()
  {
    acksNeeded++;
  }
  private synchronized void gotAck()
  {
    acksNeeded--;
    notifyAll();
  }
  private synchronized void waitAcks()
  {
    while(acksNeeded > 0) {
      try {
	wait();
      } catch(InterruptedException e) {
	System.out.println("wait() interrupted");
      }
    }
  }





  /** Causes the thread to sleep for 'amount' milliseconds.
   */
  public void sleep(int amount)
  {
    try {
      Thread.sleep(amount);
    } catch (InterruptedException e){ 
      // do nothing.
    }
  }
  
  /* Here just to satisfy my curiosity as to when this method is
     actually called. */
  protected void finalize()
  {
    System.out.println("Thread " + appNum + " to be garbage collected");
  }


  class MessageHandler
    implements MessageAction
  {

    /**
       The contents of a message should be passed in for this
       method.
    */
    public void act(Object message)
      {
	System.out.print("Message received (");

	int ident = UtilMsgr.i.getShort(message,0);
	int seq = UtilMsgr.i.getShort(message,1);
	int srcPort = UtilMsgr.i.getShort(message,2);
	int destPort = UtilMsgr.i.getShort(message,3);
	int srcIpAddr = UtilMsgr.i.getInt(message,2);
	int destIpAddr = UtilMsgr.i.getInt(message,3);

	System.out.print("ident#"+ident+" seq#"+seq+
			 " From: "+
			 IpAddress.toString(srcIpAddr)+
			 ":"+srcPort+
			 " To: "+
			 IpAddress.toString(destIpAddr)+
			 ":"+destPort);

	System.out.println(")");

	gotAck();
      }

  }

}
